@@ -159,7 +159,7 @@ Classes and functions
 ---------------------
 
 """
-from __future__ import with_statement
+from __future__ import absolute_import, division, with_statement
 
 import numpy
 import re
@@ -301,7 +301,8 @@ class array(DXclass):
         # "string" not automatically supported
     }
 
-    def __init__(self, classid, array=None, type=None, **kwargs):
+    def __init__(self, classid, array=None, type=None, typequote='"',
+                 **kwargs):
         """
         Parameters
         ----------
@@ -347,6 +348,7 @@ class array(DXclass):
                                   "types are: {1}".format(type,
                                                           list(self.dx_types.values()))))
             self.type = type
+        self.typequote = typequote
 
     def write(self, file):
         """Write the *class array* section.
@@ -367,18 +369,23 @@ class array(DXclass):
                               "Supported valus are: {}\n"
                               "Use the type=<type> keyword argument.").format(
                                   self.type, list(self.dx_types.keys())))
+        typelabel = (self.typequote+self.type+self.typequote)
         DXclass.write(self,file,
-                      'type "{0}" rank 0 items {1} data follows'.format(
-                          self.type, self.array.size))
+                      'type {0} rank 0 items {1} data follows'.format(
+                          typelabel, self.array.size))
         # grid data, serialized as a C array (z fastest varying)
         # (flat iterator is equivalent to: for x: for y: for z: grid[x,y,z])
         # VMD's DX reader requires exactly 3 values per line
+        fmt_string = "{:d}"
+        if (self.array.dtype.kind == 'f' or self.array.dtype.kind == 'c'):
+            precision = numpy.finfo(self.array.dtype).precision
+            fmt_string = "{:."+"{:d}".format(precision)+"f}"
         values_per_line = 3
         values = self.array.flat
         while 1:
             try:
                 for i in range(values_per_line):
-                    file.write(str(next(values)) + "\t")
+                    file.write(fmt_string.format(next(values)) + "\t")
                 file.write('\n')
             except StopIteration:
                 file.write('\n')
@@ -464,7 +471,7 @@ class field(DXclass):
         """
         # comments (VMD chokes on lines of len > 80, so truncate)
         maxcol = 80
-        with open(filename,'w') as outfile:
+        with open(str(filename), 'w') as outfile:
             for line in self.comments:
                 comment = '# '+str(line)
                 outfile.write(comment[:maxcol]+'\n')
@@ -605,21 +612,22 @@ class DXParser(object):
     """
 
     # the regexes must match with the categories defined in the Token class
+    # REAL regular expression will catch both integers and floats.
+    # Taken from
+    # https://docs.python.org/3/library/re.html#simulating-scanf
     dx_regex = re.compile(r"""
     (?P<COMMENT>\#.*$)            # comment (until end of line)
     |(?P<WORD>(object|class|counts|origin|delta|type|counts|rank|items|data))
     |"(?P<QUOTEDSTRING>[^\"]*)"   # string in double quotes  (quotes removed)
     |(?P<WHITESPACE>\s+)          # white space
     |(?P<REAL>[-+]?               # true real number (decimal point or
-          (\d+\.\d*([eE][-+]\d+)?)  # scientific notation)
-          |(\d*\.\d+([eE][-+]\d+)?)
-          |(\d[eE][-+]\d+))
-    |(?P<INTEGER>[-+]?\d+)       # integer
+    (\d+(\.\d*)?|\.\d+)           # scientific notation) and integers
+    ([eE][-+]?\d+)?)
     |(?P<BARESTRING>[a-zA-Z_][^\s\#\"]+) # unquoted strings, starting with non-numeric
     """, re.VERBOSE)
 
 
-    def __init__(self,filename):
+    def __init__(self, filename):
         """Setup a parser for a simple DX file (from VMD)
 
         >>> DXfield_object = OpenDX.field(id)
@@ -632,8 +640,8 @@ class DXParser(object):
 
         Note that quotes are removed from quoted strings.
         """
-        self.filename = filename
-        self.field = field('grid data',comments=['filename: '+self.filename])
+        self.filename = str(filename)
+        self.field = field('grid data',comments=['filename: {0}'.format(self.filename)])
         # other variables are initialised every time parse() is called
 
         self.parsers = {'general':self.__general,
@@ -670,7 +678,7 @@ class DXParser(object):
         self.currentobject = None           # containers for data
         self.objects = []                   # |
         self.tokens = []                    # token buffer
-        with open(self.filename,'r') as self.dxfile:
+        with open(self.filename, 'r') as self.dxfile:
             self.use_parser('general')      # parse the whole file and populate self.objects
 
         # assemble field from objects
@@ -775,10 +783,12 @@ class DXParser(object):
         if tok.equals('counts'):
             shape = []
             try:
-                while self.__peek().iscode('INTEGER'):
+                while True:
+                    # raises exception if not an int
+                    self.__peek().value('INTEGER')
                     tok = self.__consume()
-                    shape.append(tok.value())
-            except DXParserNoTokens:
+                    shape.append(tok.value('INTEGER'))
+            except (DXParserNoTokens, ValueError):
                 pass
             if len(shape) == 0:
                 raise DXParseError('gridpositions: no shape parameters')
@@ -820,7 +830,6 @@ class DXParser(object):
         pattern:
         object 2 class gridconnections counts 97 93 99
         """
-
         try:
             tok = self.__consume()
         except DXParserNoTokens:
@@ -829,10 +838,12 @@ class DXParser(object):
         if tok.equals('counts'):
             shape = []
             try:
-                while self.__peek().iscode('INTEGER'):
+                while True:
+                    # raises exception if not an int
+                    self.__peek().value('INTEGER')
                     tok = self.__consume()
-                    shape.append(tok.value())
-            except DXParserNoTokens:
+                    shape.append(tok.value('INTEGER'))
+            except (DXParserNoTokens, ValueError):
                 pass
             if len(shape) == 0:
                 raise DXParseError('gridconnections: no shape parameters')
@@ -865,16 +876,18 @@ class DXParser(object):
             self.currentobject['type'] = tok.value()
         elif tok.equals('rank'):
             tok = self.__consume()
-            if not tok.iscode('INTEGER'):
+            try:
+                self.currentobject['rank'] = tok.value('INTEGER')
+            except ValueError:
                 raise DXParseError('array: rank was "%s", not an integer.'%\
                                    tok.text)
-            self.currentobject['rank'] = tok.value()
         elif tok.equals('items'):
             tok = self.__consume()
-            if not tok.iscode('INTEGER'):
+            try:
+                self.currentobject['size'] = tok.value('INTEGER')
+            except ValueError:
                 raise DXParseError('array: items was "%s", not an integer.'%\
                                    tok.text)
-            self.currentobject['size'] = tok.value()
         elif tok.equals('data'):
             tok = self.__consume()
             if not tok.iscode('STRING'):
@@ -885,8 +898,20 @@ class DXParser(object):
                             'array: Only the "data follows header" format is supported.')
             if not self.currentobject['size']:
                 raise DXParseError("array: missing number of items")
-            self.currentobject['array'] = [self.__consume().value('REAL') \
-                                           for i in range(self.currentobject['size'])]
+            # This is the slow part.  Once we get here, we are just
+            # reading in a long list of numbers.  Conversion to floats
+            # will be done later when the numpy array is created.
+
+            # Don't assume anything about whitespace or the number of elements per row
+            self.currentobject['array'] = []
+            while len(self.currentobject['array']) <self.currentobject['size']:
+                 self.currentobject['array'].extend(self.dxfile.readline().strip().split())
+
+            # If you assume that there are three elements per row
+            # (except the last) the following version works and is a little faster.
+            # for i in range(int(numpy.ceil(self.currentobject['size']/3))):
+            #     self.currentobject['array'].append(self.dxfile.readline())
+            # self.currentobject['array'] = ' '.join(self.currentobject['array']).split()
         elif tok.equals('attribute'):
             # not used at the moment
             attribute = self.__consume().value()
