@@ -23,9 +23,16 @@ Classes and functions
 ---------------------
 
 """
+# Having consistent truedivision in this module is essential so that
+# its behavior is fully consistent in Python 2 and Python 3.
+from __future__ import absolute_import, division
+
+import six
+from six.moves import cPickle, range, zip
 
 import os
-from six.moves import cPickle, range, zip
+import errno
+
 import numpy
 
 # For interpolated grids: need scipy.ndimage but we import it only when needed:
@@ -59,26 +66,32 @@ class Grid(object):
     default_format = 'DX'
 
     def __init__(self, grid=None, edges=None, origin=None, delta=None,
-                 metadata={}, interpolation_spline_order=3):
+                 metadata={}, interpolation_spline_order=3,
+                 file_format=None):
         """
         Create a Grid object from data.
 
         From a numpy.histogramdd()::
+
           grid,edges = numpy.histogramdd(...)
           g = Grid(grid,edges=edges)
 
         From an arbitrary grid::
+
           g = Grid(grid,origin=origin,delta=delta)
 
         From a saved file::
+
           g = Grid(filename)
-        or
+
+        or ::
+
           g = Grid()
           g.load(filename)
 
         :Arguments:
           grid
-            histogram or density, defined on numpy nD array
+            histogram or density, defined on numpy nD array, or filename
           edges
             list of arrays, the lower and upper bin edges along the axes
             (both are output by numpy.histogramdd())
@@ -93,18 +106,28 @@ class Grid(object):
             metadata[] but stores it with save()
           interpolation_spline_order
             order of interpolation function for resampling; cubic splines = 3 [3]
+          file_format
+             file format; only necessary when ``grid`` is a filename (see :meth:`Grid.load`);
+             default is ``None`` and the file format is autodetected.
+
+
+        .. versionchanged:: 0.5.0
+           New *file_format* keyword argument.
+
         """
         # file formats are guess from extension == lower case key
         self._exporters = {
             'DX': self._export_dx,
-            'PICKLE': self._export_python,
+            'PKL': self._export_python,
+            'PICKLE': self._export_python,  # compatibility
             'PYTHON': self._export_python,  # compatibility
         }
         self._loaders = {
             'CCP4': self._load_cpp4,
             'DX': self._load_dx,
             'PLT': self._load_plt,
-            'PICKLE': self._load_python,
+            'PKL': self._load_python,
+            'PICKLE': self._load_python,  # compatibility
             'PYTHON': self._load_python,  # compatibility
         }
 
@@ -113,71 +136,126 @@ class Grid(object):
         self.__interpolation_spline_order = interpolation_spline_order
         self.interpolation_cval = None  # default to using min(grid)
 
-        if type(grid) is str:
-            self.load(grid)
-        elif not (grid is None or edges is None):
-            # set up from histogramdd-type data
-            self.grid = numpy.asarray(grid)
-            self.edges = edges
-            self._update()
-        elif not (grid is None or origin is None or delta is None):
-            # setup from generic data
-            origin = numpy.asarray(origin)
-            delta = numpy.asarray(delta)
-            if len(origin) != grid.ndim:
-                raise TypeError(
-                    "Dimension of origin is not the same as grid dimension.")
-            if delta.shape == () and numpy.isreal(delta):
-                delta = numpy.ones(grid.ndim) * delta
-            elif delta.ndim > 1:
-                raise NotImplementedError(
-                    "Non-rectangular grids are not supported.")
-            elif len(delta) != grid.ndim:
-                raise TypeError("delta should be scalar or array-like of"
-                                "len(grid.ndim)")
-            # note that origin is CENTER so edges must be shifted by -0.5*delta
-            self.edges = [origin[dim] +
-                          (numpy.arange(m + 1) - 0.5) * delta[dim]
-                          for dim, m in enumerate(grid.shape)]
-            self.grid = numpy.asarray(grid)
-            self._update()
-        else:
-            # empty, must manually populate with load()
-            # print "Setting up empty grid object. Use Grid.load(filename)."
-            pass
+        if grid is not None:
+            if isinstance(grid, six.string_types):
+                # can probably safely try to load() it...
+                filename = grid
+            else:
+                try:
+                    # Can we read this as a file?
+                    # Use str(x) to work with py.path.LocalPath and pathlib.Path instances
+                    # even for Python < 3.6
+                    with open(str(grid), 'rb'):
+                        pass
+                except (OSError, IOError):
+                    # no, this is probably an array-like thingy
+                    filename = None
+                else:
+                    # yes, let's use it as a file
+                    filename = str(grid)
 
-    def interpolation_spline_order():
+            if filename is not None:
+                self.load(filename, file_format=file_format)
+            else:
+                if edges is not None:
+                    # set up from histogramdd-type data
+                    self.grid = numpy.asanyarray(grid)
+                    self.edges = edges
+                    self._update()
+                elif origin is not None and delta is not None:
+                    # setup from generic data
+                    origin = numpy.asanyarray(origin)
+                    delta = numpy.asanyarray(delta)
+                    if len(origin) != grid.ndim:
+                        raise TypeError(
+                            "Dimension of origin is not the same as grid dimension.")
+                    if delta.shape == () and numpy.isreal(delta):
+                        delta = numpy.ones(grid.ndim) * delta
+                    elif delta.ndim > 1:
+                        raise NotImplementedError(
+                            "Non-rectangular grids are not supported.")
+                    elif len(delta) != grid.ndim:
+                        raise TypeError("delta should be scalar or array-like of"
+                                        "len(grid.ndim)")
+                    # note that origin is CENTER so edges must be shifted by -0.5*delta
+                    self.edges = [origin[dim] +
+                                  (numpy.arange(m + 1) - 0.5) * delta[dim]
+                                  for dim, m in enumerate(grid.shape)]
+                    self.grid = numpy.asanyarray(grid)
+                    self._update()
+                else:
+                    raise ValueError("Wrong/missing data to set up Grid. Use Grid() or "
+                                     "Grid(grid=<array>, edges=<list>) or "
+                                     "Grid(grid=<array>, origin=(x0, y0, z0), delta=(dx, dy, dz)):\n"
+                                     "grid={0} edges={1} origin={2} delta={3}".format(
+                                         grid, edges, origin, delta))
+
+    @property
+    def interpolation_spline_order(self):
         """Order of the B-spline interpolation of the data.
 
         3 = cubic; 4 & 5 are also supported
 
-        Only choose values that are acceptable to :func:`scipy.ndimage.spline_filter`!
+        Only choose values that are acceptable to
+        :func:`scipy.ndimage.spline_filter`!
+
+        See Also
+        --------
+        interpolated
         """
 
-        def fget(self):
-            return self.__interpolation_spline_order
+        return self.__interpolation_spline_order
 
-        def fset(self, x):
-            # As we cache the interpolation function, we need to rebuild the cache
-            # whenever the interpolation order changes: this is handled by _update()
-            self.__interpolation_spline_order = x
-            self._update()
+    @interpolation_spline_order.setter
+    def interpolation_spline_order(self, x):
+        """Setting the ``interpolation_spline_order`` updates :func:`interpolated`
 
-        return locals()
+        Because we cache the interpolation function, we need to rebuild the
+        cache whenever the interpolation order changes: this is
+        handled by :meth:`_update`
+
+        """
+        self.__interpolation_spline_order = x
+        self._update()
 
-    interpolation_spline_order = property(**interpolation_spline_order())
 
     def resample(self, edges):
         """Resample data to a new grid with edges *edges*.
 
-          resample(edges) --> Grid
+        This method creates a new grid with the data from the current
+        grid resampled to a regular grid specified by *edges*.  The
+        order of the interpolation is set by
+        :attr:`Grid.interpolation_spline_order`: change the value
+        *before* calling :meth:`resample`.
 
-        or
+        Parameters
+        ----------
+        edges : tuple of arrays or Grid
+             edges of the new grid or a :class:`Grid` instance that
+             provides :attr:`Grid.edges`
 
-          resample(otherGrid) --> Grid
+        Returns
+        -------
+        Grid
+             a new :class:`Grid` with the data interpolated over the
+             new grid cells
+
+
+        Examples
+        --------
+
+        Providing *edges* (a tuple of three arrays, indicating the
+        boundaries of each grid cell)::
+
+          g = grid.resample(edges)
+
+        As a convenience, one can also supply another :class:`Grid` as
+        the argument for this method ::
+
+          g = grid.resample(othergrid)
+
+        and the edges are taken from :attr:`Grid.edges`.
 
-        The order of the interpolation is set by
-        :attr:`Grid.interpolation_spline_order`.
         """
         try:
             edges = edges.edges  # can also supply another Grid
@@ -187,14 +265,33 @@ class Grid(object):
         coordinates = ndmeshgrid(*midpoints)
         # feed a meshgrid to generate all points
         newgrid = self.interpolated(*coordinates)
-        return Grid(newgrid, edges)
+        return self.__class__(newgrid, edges)
 
     def resample_factor(self, factor):
-        """Resample to a new regular grid with factor*oldN cells along
-        each dimension."""
+        """Resample to a new regular grid.
+
+
+        Parameters
+        ----------
+        factor : float
+            The number of grid cells are scaled with `factor` in each
+            dimension, i.e., ``factor * N_i`` cells along each
+            dimension i.
+
+
+        Returns
+        -------
+        Grid
+
+
+        See Also
+        --------
+        resample
+
+        """
         # new number of edges N' = (N-1)*f + 1
         newlengths = [(N - 1) * float(factor) + 1 for N in self._len_edges()]
-        edges = [numpy.linspace(start, stop, num=N, endpoint=True)
+        edges = [numpy.linspace(start, stop, num=int(N), endpoint=True)
                  for (start, stop, N) in
                  zip(self._min_edges(), self._max_edges(), newlengths)]
         return self.resample(edges)
@@ -202,8 +299,6 @@ class Grid(object):
     def _update(self):
         """compute/update all derived data
 
-        Grid._update()
-
         Can be called without harm and is idem-potent.
 
         Updates these attributes and methods:
@@ -227,20 +322,36 @@ class Grid(object):
     def interpolated(self):
         """B-spline function over the data grid(x,y,z).
 
+        The :func:`interpolated` function allows one to obtain data
+        values for any values of the coordinates::
+
            interpolated([x1,x2,...],[y1,y2,...],[z1,z2,...]) -> F[x1,y1,z1],F[x2,y2,z2],...
 
-        The interpolation order is set in :attr:`Grid.interpolation_spline_order`.
+        The interpolation order is set in
+        :attr:`Grid.interpolation_spline_order`.
 
         The interpolated function is computed once and is cached for better
         performance. Whenever :attr:`~Grid.interpolation_spline_order` is
         modified, :meth:`Grid.interpolated` is recomputed.
 
         The value for unknown data is set in :attr:`Grid.interpolation_cval`
-        (TODO: also recompute when interpolation_cval value is changed.)
+        (TODO: also recompute when ``interpolation_cval`` value is changed.)
 
+        Example
+        -------
         Example usage for resampling::
-           >>> XX,YY,ZZ = numpy.mgrid[40:75:0.5, 96:150:0.5, 20:50:0.5]
-           >>> FF = interpolated(XX,YY,ZZ)
+
+           XX, YY, ZZ = numpy.mgrid[40:75:0.5, 96:150:0.5, 20:50:0.5]
+           FF = interpolated(XX, YY, ZZ)
+
+        Note
+        ----
+        Values are interpolated with a spline function. It is possible
+        that the spline will generate values that would not normally
+        appear in the data. For example, a density is non-negative but
+        a cubic spline interpolation can generate negative values,
+        especially at the boundary between 0 and high values.
+
         """
         if self.__interpolated is None:
             self.__interpolated = self._interpolationFunctionFactory()
@@ -298,6 +409,12 @@ class Grid(object):
         The load() method calls the class's constructor method and
         completely resets all values, based on the loaded data.
         """
+        filename = str(filename)
+        if not os.path.exists(filename):
+            # check before we try to detect the file type because
+            # _guess_fileformat() does not work well with things that
+            # are not really a file
+            raise IOError(errno.ENOENT, "file not found", filename)
         loader = self._get_loader(filename, file_format=file_format)
         loader(filename)
 
@@ -329,7 +446,7 @@ class Grid(object):
         grid, edges = g.histogramdd()
         self.__init__(grid=grid, edges=edges, metadata=self.metadata)
 
-    def export(self, filename, file_format=None, type=None):
+    def export(self, filename, file_format=None, type=None, typequote='"'):
         """export density to file using the given format.
 
         The format can also be deduced from the suffix of the filename
@@ -343,30 +460,38 @@ class Grid(object):
         dx
             :mod:`OpenDX`
         pickle
-            pickle (use :meth:``Grid.load` to restore); :meth:`Grid.save`
+            pickle (use :meth:`Grid.load` to restore); :meth:`Grid.save`
             is simpler than ``export(format='python')``.
 
         Parameters
         ----------
+
         filename : str
             name of the output file
+
         file_format : {'dx', 'pickle', None} (optional)
             output file format, the default is "dx"
+
         type : str (optional)
-            for DX, set the output DX array type, e.g., "double" or
-            "float"; note that PyMOL only understands "double" (see
-            issue `#35`_). By default (``None``), the DX type is
-            determined from the numpy dtype of the array of the grid
-            (and this will typically result in "double").
+            for DX, set the output DX array type, e.g., "double" or "float".
+            By default (``None``), the DX type is determined from the numpy
+            dtype of the array of the grid (and this will typically result in
+            "double").
 
             .. versionadded:: 0.4.0
 
+        typequote : str (optional)
+            For DX, set the character used to quote the type string;
+            by default this is a double-quote character, '"'.
+            Custom parsers like the one from NAMD-GridForces (backend for MDFF)
+            expect no quotes, and typequote='' may be used to appease them.
 
-        .. _`#35`: https://github.com/MDAnalysis/GridDataFormats/issues/35
+            .. versionadded:: 0.5.0
 
         """
+        filename = str(filename)
         exporter = self._get_exporter(filename, file_format=file_format)
-        exporter(filename, type=type)
+        exporter(filename, type=type, typequote=typequote)
 
     # note: the _export_FORMAT() methods all take the filename as a mandatory
     # argument. They can process kwargs but they are not required to do
@@ -378,21 +503,20 @@ class Grid(object):
         The object is dumped as a dictionary with grid and edges: This
         is sufficient to recreate the grid object with __init__().
         """
-        root, ext = os.path.splitext(filename)
-        filename = root + ".pickle"
-
         data = dict(grid=self.grid, edges=self.edges, metadata=self.metadata)
         with open(filename, 'wb') as f:
             cPickle.dump(data, f, cPickle.HIGHEST_PROTOCOL)
 
-    def _export_dx(self, filename, type=None, **kwargs):
-        """Export the density grid to an OpenDX file. The file format
-        is the simplest regular grid array and it is also understood
-        by VMD's and Chimera's DX reader; PyMOL requires the dx `type`
-        to be set to "double".
+    def _export_dx(self, filename, type=None, typequote='"', **kwargs):
+        """Export the density grid to an OpenDX file.
+
+        The file format is the simplest regular grid array and it is
+        also understood by VMD's and Chimera's DX reader; PyMOL
+        requires the dx `type` to be set to "double".
 
         For the file format see
         http://opendx.sdsc.edu/docs/html/pages/usrgu068.htm#HDREDF
+
         """
         root, ext = os.path.splitext(filename)
         filename = root + '.dx'
@@ -417,7 +541,7 @@ class Grid(object):
             positions=OpenDX.gridpositions(1, self.grid.shape, self.origin,
                                            self.delta),
             connections=OpenDX.gridconnections(2, self.grid.shape),
-            data=OpenDX.array(3, self.grid, type=type),
+            data=OpenDX.array(3, self.grid, type=type, typequote=typequote),
         )
         dx = OpenDX.field('density', components=components, comments=comments)
         dx.write(filename)
@@ -425,12 +549,17 @@ class Grid(object):
     def save(self, filename):
         """Save a grid object to <filename>.pickle
 
-           Grid.save(filename)
+        Internally, this calls
+        ``Grid.export(filename, format="python")``. A grid can be
+        regenerated from the saved data with ::
 
-        Internally, this calls Grid.export(filename,format="python"). A grid can be
-        regenerated from the saved data with
+           g = Grid(filename="grid.pickle")
 
-           g = Grid(filename=<filename>)
+        .. note::
+           The pickle format depends on the Python version and
+           therefore it is not guaranteed that a grid saved with, say,
+           Python 2.7 can also be read with Python 3.5. The OpenDX format
+           is a better alternative for portability.
 
         """
         self.export(filename, file_format="pickle")
@@ -520,65 +649,86 @@ class Grid(object):
     def __eq__(self, other):
         if not isinstance(other, Grid):
             return False
-        return numpy.all(other.grid == self.grid) and numpy.all(
-            other.origin == self.origin) and other.edges == self.edges
+        return numpy.all(other.grid == self.grid) and \
+            numpy.all(other.origin == self.origin) and \
+            numpy.all(numpy.all(other_edge == self_edge) for other_edge, self_edge in
+                      zip(other.edges, self.edges))
 
     def __ne__(self, other):
         return not self.__eq__(other)
 
     def __add__(self, other):
         self.check_compatible(other)
-        return Grid(self.grid + _grid(other), edges=self.edges)
+        return self.__class__(self.grid + _grid(other), edges=self.edges)
 
     def __sub__(self, other):
         self.check_compatible(other)
-        return Grid(self.grid - _grid(other), edges=self.edges)
+        return self.__class__(self.grid - _grid(other), edges=self.edges)
 
     def __mul__(self, other):
         self.check_compatible(other)
-        return Grid(self.grid * _grid(other), edges=self.edges)
-
-    def __div__(self, other):
-        self.check_compatible(other)
-        return Grid(self.grid / _grid(other), edges=self.edges)
+        return self.__class__(self.grid * _grid(other), edges=self.edges)
 
     def __truediv__(self, other):
+        # truediv will always do true division (in Python 2 and Python 3);
+        # we use from __future__ include division everywhere
         self.check_compatible(other)
-        return Grid(self.grid / _grid(other), edges=self.edges)
+        return self.__class__(self.grid / _grid(other), edges=self.edges)
+
+    def __div__(self, other):
+        # in Python 2 only (without __future__.division): will do "classic division"
+        # https://docs.python.org/2/reference/datamodel.html#object.__div__
+        if not six.PY2:
+            raise NotImplementedError("__div__ is only available in Python 2, use __truediv__")
+        self.check_compatible(other)
+        return self.__class__(self.grid.__div__(_grid(other)), edges=self.edges)
+
+    def __floordiv__(self, other):
+        self.check_compatible(other)
+        return self.__class__(self.grid // _grid(other), edges=self.edges)
 
     def __pow__(self, other):
         self.check_compatible(other)
-        return Grid(numpy.power(self.grid, _grid(other)), edges=self.edges)
+        return self.__class__(numpy.power(self.grid, _grid(other)), edges=self.edges)
 
     def __radd__(self, other):
         self.check_compatible(other)
-        return Grid(_grid(other) + self.grid, edges=self.edges)
+        return self.__class__(_grid(other) + self.grid, edges=self.edges)
 
     def __rsub__(self, other):
         self.check_compatible(other)
-        return Grid(_grid(other) - self.grid, edges=self.edges)
+        return self.__class__(_grid(other) - self.grid, edges=self.edges)
 
     def __rmul__(self, other):
         self.check_compatible(other)
-        return Grid(_grid(other) * self.grid, edges=self.edges)
-
-    def __rdiv__(self, other):
-        self.check_compatible(other)
-        return Grid(_grid(other) / self.grid, edges=self.edges)
+        return self.__class__(_grid(other) * self.grid, edges=self.edges)
 
     def __rtruediv__(self, other):
-        return Grid(_grid(other) / self.grid, edges=self.edges)
+        self.check_compatible(other)
+        return self.__class__(_grid(other) / self.grid, edges=self.edges)
+
+    def __rdiv__(self, other):
+        # in Python 2 only (without __future__.division): will do "classic division"
+        # https://docs.python.org/2/reference/datamodel.html#object.__div__
+        if not six.PY2:
+            raise NotImplementedError("__rdiv__ is only available in Python 2, use __rtruediv__")
+        self.check_compatible(other)
+        return self.__class__(self.grid.__rdiv__(_grid(other)), edges=self.edges)
+
+    def __rfloordiv__(self, other):
+        self.check_compatible(other)
+        return self.__class__(_grid(other) // self.grid, edges=self.edges)
 
     def __rpow__(self, other):
         self.check_compatible(other)
-        return Grid(numpy.power(_grid(other), self.grid), edges=self.edges)
+        return self.__class__(numpy.power(_grid(other), self.grid), edges=self.edges)
 
     def __repr__(self):
         try:
             bins = self.grid.shape
         except AttributeError:
             bins = "no"
-        return '<Grid with ' + str(bins) + ' bins>'
+        return '<{0} with {1!r} bins>'.format(self.__class__, bins)
 
 
 def ndmeshgrid(*arrs):
@@ -607,7 +757,7 @@ def ndmeshgrid(*arrs):
     for i, arr in enumerate(arrs):
         slc = [1] * dim
         slc[i] = lens[i]
-        arr2 = numpy.asarray(arr).reshape(slc)
+        arr2 = numpy.asanyarray(arr).reshape(slc)
         for j, sz in enumerate(lens):
             if j != i:
                 arr2 = arr2.repeat(sz, axis=j)
